// file      : xsd/cxx/parser/parser-inline.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/parser-inline.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      //
      //
      struct Member: Traversal::Member,
                     protected virtual Context
      {
        Member (Context& c)
            : Context (c),
              template_param_decl_list_ (c),
              template_param_use_list_ (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          String class_name (id (m.scope ().name ()));

          enter_class_scope (class_name, m.scope ());

          String name (id (m.name ()));
          String type (id ("_xsd_", m.name ()));
          String inst (id ("_xsd_", m.name (), "_"));

          leave_scope ();

          os <<"// " << class_name << "::" << name << endl
             <<"//" << endl
             << endl;


          // name (traverser<type>&)
          //
          os << "template < ";

          template_param_decl_list_.dispatch (m.scope ());

          os << " >" << endl
             << inl
             << "void " << class_name << "< ";

          template_param_use_list_.dispatch (m.scope ());

          os << " >::" << endl
             << name << " (" << parser_template << "< " << type << ", " <<
            char_type << " >& " << name << ")"
             << "{"
             << "this->" << inst << " = &" << name << ";"
             << "}"
             << endl;
        }

      private:
        TemplateParamDeclList template_param_decl_list_;
        TemplateParamUseList template_param_use_list_;
      };


      //
      //
      struct ParserMemberSet : Traversal::Member,
                               protected virtual Context
      {
        ParserMemberSet (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          String name (m.name ());

          os << "this->" << id ("_xsd_", name, "_")  << " = &" <<
            id (name) << ";";
        }
      };


      //
      //
      struct ParserBaseSet : Traversal::Complex,
                             protected virtual Context
      {
        ParserBaseSet (Context& c)
            : Context (c), member_ (c)
        {
          *this >> inherits_ >> *this;
          *this >> names_ >> member_;
        }

        virtual Void
        traverse (Type& c)
        {
          Traversal::Complex::traverse (c);
        }

      private:
        Traversal::Inherits inherits_;

        ParserMemberSet member_;
        Traversal::Names names_;
      };


      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              member_ (c),
              parser_member_set_ (c),
              parser_base_set_ (c),
              template_param_decl_list_ (c),
              template_param_use_list_ (c)
        {
          names_member_ >> member_;

          parser_set_names_ >> parser_member_set_;
          parser_set_inherits_ >> parser_base_set_;
        }

        virtual Void
        traverse (Type& c)
        {
          String name (id (c.name ()));
          String type (id ("_xsd_", c.name ()));

          if (!(has<Traversal::Element> (c) || has<Traversal::Attribute> (c)))
              return;


          os << "// " << name << endl
             << "//" << endl
             << endl;

          // operator ()
          //

          os << "template < ";

          {
            template_param_decl_list_.traverse (c);
          }

          os << " >" << endl
             << inl
             << "void " << name << "< ";

          {
            template_param_use_list_.traverse (c);
          }

          os << " >::" << endl
             << "operator() (";

          {
            ParserParamDecl decl (*this, true);
            decl.traverse (c);
          }

          os << ")"
             << "{";

          {
            enter_class_scope (name, c);
            inherits (c, parser_set_inherits_);
            names (c, parser_set_names_);
            leave_scope ();
          }

          os << "}"
             << endl;

          {
            names (c, names_member_);
          }
        }

      private:
        //
        //
        Member member_;
        Traversal::Names names_member_;

        //
        //
        ParserMemberSet parser_member_set_;
        Traversal::Names parser_set_names_;

        ParserBaseSet parser_base_set_;
        Traversal::Inherits parser_set_inherits_;

        //
        //
        TemplateParamDeclList template_param_decl_list_;
        TemplateParamUseList template_param_use_list_;
      };
    }

    Void
    generate_parser_inline (Context& ctx, SemanticGraph::Schema& tu)
    {
      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      Complex complex (ctx);

      names >> complex;

      schema.dispatch (tu);
    }
  }
}

