// file      : xsd/cxx/parser/parser-template.cxx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005-2006 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <cxx/parser/parser-template.hxx>

#include <xsd-frontend/semantic-graph.hxx>
#include <xsd-frontend/traversal.hxx>

#include <backend-elements/regex.hxx>

namespace CXX
{
  namespace Parser
  {
    namespace
    {
      //
      //
      struct Member: Traversal::Member,
                     protected virtual Context
      {
        Member (Context& c)
            : Context (c),
              template_param_decl_list_ (c),
              template_param_use_list_ (c)
        {
        }

        virtual Void
        traverse (Type& m)
        {
          String class_name (id (m.scope ().name ()));

          enter_class_scope (class_name, m.scope ());

          String name (id (m.name ()));
          String type (id ("_xsd_", m.name ()));

          leave_scope ();

          os <<"// " << class_name << "::" << name << endl
             <<"//" << endl
             << endl;

          // name (type const&)
          //
          os << "template < ";

          template_param_decl_list_.dispatch (m.scope ());

          os << " >" << endl
             << "void " << class_name << "< ";

          template_param_use_list_.dispatch (m.scope ());

          os << " >::" << endl
             << name << " (" << type << " const&)"
             << "{"
             << "}";
        }

      private:
        TemplateParamDeclList template_param_decl_list_;
        TemplateParamUseList template_param_use_list_;
      };


      //
      //
      struct StartElement : Traversal::Element,
                            protected virtual Context
      {
        StartElement (Context& c, String const& type_)
            : Context (c), type (type_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String inst (id ("_xsd_", e.name (), "_"));


          if (e.qualified ())
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && " <<
              "ns == " << L << "\"" << e.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && ns.empty ())"
               << "{";
          }

          os << "if (" << parser_template << "< " << type << ", " <<
            char_type << " >::_start_element_ (ns, n, " << inst << ")) " <<
            inst << "->pre ();"
             << "}"
             << "else ";
        }

      private:
        String type;
      };


      //
      //
      struct EndElement : Traversal::Element,
                          protected virtual Context
      {
        EndElement (Context& c, String const& type_)
            : Context (c), type (type_)
        {
        }

        virtual Void
        traverse (Type& e)
        {
          String name (id (e.name ()));
          String inst (id ("_xsd_", e.name (), "_"));


          if (e.qualified ())
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && " <<
              "ns == " << L << "\"" << e.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << e.name () << "\" && ns.empty ())"
               << "{";
          }

          os << "if (" << parser_template << "< " << type << ", " <<
            char_type << " >::_end_element_ (ns, n)) " <<
            name << " (" << inst << "->post ());"
             << "}"
             << "else ";
        }

      private:
        String type;
      };


      //
      //
      struct Attribute : Traversal::Attribute,
                         protected virtual Context
      {
        Attribute (Context& c)
            : Context (c)
        {
        }

        virtual Void
        traverse (Type& a)
        {
          String name (id (a.name ()));
          String inst (id ("_xsd_", a.name (), "_"));

          if (a.qualified ())
          {
            os << "if (n == " << L << "\"" << a.name () << "\" && " <<
              "ns == " << L << "\"" << a.namespace_ ().name () << "\")"
               << "{";
          }
          else
          {
            os << "if (n == " << L << "\"" << a.name () << "\" && ns.empty ())"
               << "{";
          }

          os << inst << "->pre ();"
             << inst << "->_characters (v);"
             << name << " (" << inst << "->post ());"
             << "}"
             << "else ";
        }
      };


      //
      //
      struct BaseType: Base,
                       Traversal::Type
      {
        BaseType (Context& c, String const& name_)
            : Context (c), Base (c), name (name_)
        {
        }

        virtual Void
        traverse (SemanticGraph::Type& t)
        {
          os << parser_template << "< " << name << ", " <<
            char_type << " >";
        }

      private:
        String name;
      };


      //
      //
      struct Complex : Traversal::Complex,
                       protected virtual Context
      {
        Complex (Context& c)
            : Context (c),
              member_ (c),
              attribute_ (c),
              template_param_decl_list_ (c),
              template_param_use_list_ (c)
        {
          names_member_ >> member_;

          attribute_names_ >> attribute_;
        }

        virtual Void
        traverse (Type& c)
        {
          String name (id (c.name ()));
          String type (id ("_xsd_", c.name ()));

          os << "// " << name << endl
             << "//" << endl
             << endl;

          //
          //
          BaseType base_type_ (*this, type);
          Traversal::Inherits inherits_type_;
          inherits_type_ >> base_type_;


          // start_element
          //
          os << "template < ";

          {
            template_param_decl_list_.traverse (c);
          }

          os << " >" << endl
             << "void " << name << "< ";

          {
            template_param_use_list_.traverse (c);
          }

          os << " >::" << endl
             << "_start_element (" <<
            string_type << " const& ns, " <<
            string_type << " const& n)"
             << "{";

          {
            StartElement start_element_ (*this, type);
            Traversal::Names start_element_names_;

            start_element_names_ >> start_element_;

            enter_class_scope (name, c);
            names (c, start_element_names_);
            leave_scope ();
          }

          os << "{";

          inherits<Complex> (c, inherits_type_, 0, 0, &Complex::default_base);

          os << "::" << "_start_element (ns, n);"
             << "}"
             << "}";

          // end_element
          //
          os << "template < ";

          {
            template_param_decl_list_.traverse (c);
          }

          os << " >" << endl
             << "void " << name << "< ";

          {
            template_param_use_list_.traverse (c);
          }


          os << " >::" << endl
             << "_end_element (" <<
            string_type << " const& ns, " <<
            string_type << " const& n)"
             << "{";

          {
            EndElement end_element_ (*this, type);
            Traversal::Names end_element_names_;

            end_element_names_ >> end_element_;

            enter_class_scope (name, c);
            names (c, end_element_names_);
            leave_scope ();
          }

          os << "{";

          inherits<Complex> (c, inherits_type_, 0, 0, &Complex::default_base);

          os << "::" << "_end_element (ns, n);"
             << "}"
             << "}";


          // attribute
          //
          os << "template < ";

          {
            template_param_decl_list_.traverse (c);
          }

          os << " >" << endl
             << "void " << name << "< ";

          {
            template_param_use_list_.traverse (c);
          }

          os << " >::" << endl
             << "_attribute (" <<
            string_type << " const& ns, " <<
            string_type << " const& n, " <<
            string_type << " const& v)"
             << "{";

          {
            enter_class_scope (name, c);
            names (c, attribute_names_);
            leave_scope ();
          }

          os << "{";

          inherits<Complex> (c, inherits_type_, 0, 0, &Complex::default_base);

          os << "::" << "_attribute (ns, n, v);"
             << "}"
             << "}"
             << endl;

          {
            names (c, names_member_);
          }
        }

        virtual Void
        default_base (Type& c)
        {
          os << parser_template << "< " << id ("_xsd_", c.name ()) << ", " <<
            char_type << " >";
        }

      private:
        //
        //
        Member member_;
        Traversal::Names names_member_;

        //
        //
        Attribute attribute_;
        Traversal::Names attribute_names_;

        //
        //
        TemplateParamDeclList template_param_decl_list_;
        TemplateParamUseList template_param_use_list_;
      };
    }

    Void
    generate_parser_template (Context& ctx, SemanticGraph::Schema& tu)
    {
      Traversal::Schema schema;

      Traversal::Sources sources;
      Traversal::Names schema_names;

      Namespace ns (ctx);
      Traversal::Names names;

      schema >> sources >> schema;
      schema >> schema_names >> ns >> names;

      Complex complex (ctx);

      names >> complex;

      schema.dispatch (tu);
    }
  }
}

